import collections
import enum

from unittest import TestCase

from django.core.exceptions import ImproperlyConfigured
from rest_framework import fields
from rest_framework.exceptions import ValidationError
from rest_framework.serializers import Serializer

from rest_framework_dataclasses.fields import DefaultDecimalField, EnumField, IterableField, MappingField, UnionField


class FieldTest(TestCase):
    def test_decimal_field(self):
        # check no parameters are required (this shouldn't throw)
        field = DefaultDecimalField()

        # check no parameters are overwritten
        field = DefaultDecimalField(max_digits=4, decimal_places=6)
        self.assertEqual(field.max_digits, 4)
        self.assertEqual(field.decimal_places, 6)

    def test_enum_field(self):
        class Color(enum.Enum):
            RED = 'FF0000'
            GREEN = '00FF00'
            BLUE = '0000FF'

        field = EnumField(Color)
        self.assertDictEqual(field.choices, {'FF0000': 'RED', '00FF00': 'GREEN', '0000FF': 'BLUE'})
        self.assertEqual(field.to_representation(Color.GREEN), '00FF00')
        self.assertEqual(field.to_internal_value('00FF00'), Color.GREEN)
        with self.assertRaises(ValidationError):
            field.to_internal_value('RED')

        field = EnumField(Color, by_name=True)
        self.assertDictEqual(field.choices, {'RED': 'RED', 'GREEN': 'GREEN', 'BLUE': 'BLUE'})
        self.assertEqual(field.to_internal_value('GREEN'), Color.GREEN)
        self.assertEqual(field.to_representation(Color.GREEN), 'GREEN')
        with self.assertRaises(ValidationError):
            field.to_internal_value('FF0000')

        self.assertEqual(field.to_representation('RED'), 'RED')
        with self.assertRaises(ValidationError):
            field.to_representation('FFFFFF')

        # check explicit specification of options
        field = EnumField(Color, choices=[('FF0000', 'RED'), ('00FF00', 'GREEN')])
        self.assertEqual(len(field.choices), 2)

    def test_iterable_field(self):
        default_field = IterableField()
        self.assertEqual(default_field.to_internal_value(['foo', 'bar']), ['foo', 'bar'])

        set_field = IterableField(container=set)
        self.assertEqual(set_field.to_internal_value(['foo', 'bar', 'baz']), {'foo', 'bar', 'baz'})

    def test_mapping_field(self):
        default_field = MappingField()
        self.assertEqual(default_field.to_internal_value({'foo': 'bar'}), {'foo': 'bar'})

        ordered_field = MappingField(container=collections.OrderedDict)
        ordered_values = {'foo': 'bar', 'abc': 'def'}
        self.assertEqual(ordered_field.to_internal_value(ordered_values), collections.OrderedDict(ordered_values))

    def test_union_field(self):
        class A:
            def __init__(self, a):
                self.a = a

        class B:
            def __init__(self, b):
                self.b = b

        class ASerializer(Serializer):
            a = fields.CharField()

        class BSerializer(Serializer):
            b = fields.IntegerField()

        # Note that A/B are regular classes and serializers, so internal value is a dict and not a dataclass.
        ab_field = UnionField({A: ASerializer, B: BSerializer}, nest_value=False)
        self.assertEqual(ab_field.to_internal_value({'type': 'A', 'a': 'a'}), {'a': 'a'})
        self.assertEqual(ab_field.to_internal_value({'type': 'B', 'b': 42}), {'b': 42})
        self.assertEqual(ab_field.to_representation(A('a')), {'type': 'A', 'a': 'a'})
        self.assertEqual(ab_field.to_representation(B(42)), {'type': 'B', 'b': 42})

        strint_field = UnionField({str: fields.CharField, int: fields.IntegerField}, nest_value=True)
        self.assertEqual(strint_field.to_internal_value({'type': 'str', 'value': 'hello'}), 'hello')
        self.assertEqual(strint_field.to_internal_value({'type': 'int', 'value': 42}), 42)
        self.assertEqual(strint_field.to_representation('hello'), {'type': 'str', 'value': 'hello'})
        self.assertEqual(strint_field.to_representation(42), {'type': 'int', 'value': 42})

        with self.assertRaises(ValidationError):
            strint_field.to_internal_value({'value': 3})
        with self.assertRaises(ValidationError):
            strint_field.to_internal_value({'type': 'int'})
        with self.assertRaises(ValidationError):
            strint_field.to_internal_value({'type': 'int', 'value': 3.5})
        with self.assertRaises(ValidationError):
            strint_field.to_internal_value({'type': 'float', 'value': 3})

        invalid_field = UnionField({str: fields.CharField, int: fields.IntegerField}, nest_value=False)
        with self.assertRaises(ImproperlyConfigured):
            invalid_field.to_representation(42)

        renamed_field = UnionField({int: fields.IntegerField, float: fields.FloatField},
                                   nest_value=True,
                                   discriminator_field_name='renamed_type',
                                   value_field_name='renamed_value')
        self.assertEqual(renamed_field.to_internal_value({'renamed_type': 'float', 'renamed_value': 1.2}), 1.2)
        self.assertEqual(renamed_field.to_representation(3.4), {'renamed_type': 'float', 'renamed_value': 3.4})



import collections.abc
from typing import Any, Dict, Generic, Type, TypeVar, Union, Optional

from django.core.exceptions import ImproperlyConfigured
from rest_framework.exceptions import ValidationError
from rest_framework.fields import Field, ChoiceField, DecimalField, DictField, ListField
from rest_framework_dataclasses import field_utils

T = TypeVar('T')


class DefaultDecimalField(DecimalField):
    def __init__(self, **kwargs):
        if 'max_digits' not in kwargs:
            kwargs['max_digits'] = None
        if 'decimal_places' not in kwargs:
            # Maybe this should be configurable, but it doesn't seem that useful. File an issue if you want it to.
            kwargs['decimal_places'] = 2

        super(DefaultDecimalField, self).__init__(**kwargs)


class EnumField(ChoiceField):
    def __init__(self, enum_class, by_name=False, **kwargs):
        self.enum_class = enum_class
        self.by_name = by_name
        if 'choices' not in kwargs:
            kwargs['choices'] = [(self.to_representation(member), member.name) for member in self.enum_class]

        super(EnumField, self).__init__(**kwargs)

    def to_internal_value(self, data):
        try:
            if self.by_name:
                return self.enum_class[data]
            else:
                return self.enum_class(data)
        except (KeyError, ValueError):
            self.fail('invalid_choice', input=data)

    def to_representation(self, value):
        # Some external libraries expect to be able to call to_representation() with the key from the choices
        # array, which seems at least somewhat reasonable. See #40.
        if not isinstance(value, self.enum_class):
            if value in self.choices:
                return value
            self.fail('invalid_choice', input=value)

        if self.by_name:
            return value.name
        else:
            return value.value


class IterableField(ListField):
    def __init__(self, **kwargs):
        self.container = kwargs.pop('container', list)
        super(IterableField, self).__init__(**kwargs)

    def to_internal_value(self, value):
        return self.container(super(IterableField, self).to_internal_value(value))


class MappingField(DictField):
    def __init__(self, **kwargs):
        self.container = kwargs.pop('container', dict)
        super(MappingField, self).__init__(**kwargs)

    def to_internal_value(self, value):
        return self.container(super(MappingField, self).to_internal_value(value))


class UnionField(Field, Generic[T]):
    # Name of the field with the tag that identifies which of the member types is contained in the field.
    discriminator_field_name = 'type'
    # Name for `nest_value` fields.
    value_field_name = 'value'

    child_fields: Dict[str, Field]
    type_mapping: Dict[type, str]

    def __init__(self,
                 child_fields: Dict[type, Union[Field, Type[Field]]],
                 *,
                 nest_value: bool = False,
                 discriminator_field_name: Optional[str] = None,
                 value_field_name: Optional[str] = None,
                 **kwargs: Any):
        super().__init__(**kwargs)
        self.nest_value = nest_value
        if discriminator_field_name is not None:
            self.discriminator_field_name = discriminator_field_name
        if value_field_name is not None:
            self.value_field_name = value_field_name

        self.child_fields = {}
        self.type_mapping = {}
        for tp, field in child_fields.items():
            discriminator = self.get_discriminator(tp)
            if isinstance(field, type):
                field = field()
            self.child_fields[discriminator] = field
            self.type_mapping[tp] = discriminator

    def get_discriminator(self, tp: type) -> str:
        return tp.__name__

    def to_representation(self, data: T) -> Dict[str, Any]:
        discriminator = field_utils.lookup_type_in_mapping(self.type_mapping, type(data))
        field = self.child_fields[discriminator]

        ret = field.to_representation(data)
        if self.nest_value:
            ret = {self.value_field_name: ret}
        elif not isinstance(ret, collections.abc.Mapping):
            raise ImproperlyConfigured(
                'Representation for `{type}` value of `{serializer}.{field}` is not a mapping, which prohibits '
                'insertion of discriminator field. Either enable `nest_value` on the `UnionField`, or change the '
                '`{type}` field to one that returns a mapping.'.format(
                    type=type(data).__name__,
                    serializer=self.parent.__class__.__name__,
                    field=self.field_name,
                )
            )

        ret[self.discriminator_field_name] = discriminator
        return ret

    def to_internal_value(self, data: Dict[str, Any]) -> T:
        if self.discriminator_field_name not in data:
            raise ValidationError({self.discriminator_field_name: 'Discriminator field must be present.'})
        discriminator = data[self.discriminator_field_name]
        if discriminator not in self.child_fields:
            raise ValidationError({self.discriminator_field_name: 'Not a valid type.'})
        field = self.child_fields[discriminator]

        if self.nest_value:
            if self.value_field_name not in data:
                raise ValidationError({self.value_field_name: 'This field is required.'})
            data = data[self.value_field_name]

        return field.to_internal_value(data)



import unittest
import typing

from rest_framework_dataclasses import field_utils
from rest_framework_dataclasses.types import Literal


class FieldUtilsTest(unittest.TestCase):
    def test_lookup_type(self):
        mapping = {
            str: 1,
            int: 2,
            typing.Union[int, str]: 3
        }

        self.assertEqual(field_utils.lookup_type_in_mapping(mapping, str), 1)
        self.assertEqual(field_utils.lookup_type_in_mapping(mapping, int), 2)
        self.assertEqual(field_utils.lookup_type_in_mapping(mapping, typing.Union[int, str]), 3)

        self.assertEqual(field_utils.lookup_type_in_mapping(mapping, type('email', (str, ), {})), 1)

        with self.assertRaises(KeyError):
            field_utils.lookup_type_in_mapping(mapping, float)
        with self.assertRaises(KeyError):
            field_utils.lookup_type_in_mapping(mapping, Literal['a', 'b'])



import collections
import dataclasses
import datetime
import decimal
import enum
import re
import sys
import typing
import unittest
import uuid
from collections import abc

import django
from django.db import models
from rest_framework import fields, relations

from rest_framework_dataclasses import field_utils, fields as custom_fields
from rest_framework_dataclasses.serializers import DataclassSerializer
from rest_framework_dataclasses.types import Final, Literal


class FieldsTest(unittest.TestCase):
    def build_typed_field(self, type_hint, field=None, extra_kwargs=None):
        field_tuple = ('test_field', type_hint, field) if field else ('test_field', type_hint)
        testclass = dataclasses.make_dataclass('TestDataclass', [field_tuple])
        serializer = DataclassSerializer(dataclass=testclass)
        type_info = field_utils.get_type_info(serializer.dataclass_definition.field_types['test_field'])

        extra_kwargs = extra_kwargs or {}
        return serializer.build_typed_field('test_field', type_info, extra_kwargs)

    def check_field_additional(self, type_hint, field, serializer_field, serializer_kwargs=None):
        field_class, field_kwargs = self.build_typed_field(type_hint, field)
        self.assertEqual(field_class, serializer_field)

        if serializer_kwargs is not None:
            if 'child' in field_kwargs:
                field_kwargs['child'] = type(field_kwargs['child'])
            self.assertDictEqual(field_kwargs, serializer_kwargs, f'arguments for field of type {type_hint}')

    def check_field(self, type_hint, field, kwargs=None):
        self.check_field_additional(type_hint, None, field, kwargs)

    def test_arguments(self):
        # Check that no qualifiers emit no arguments.
        self.check_field_additional(int, None, fields.IntegerField, {})

        # Check that a default value (factory) sets required to False (#16, #38).
        self.check_field_additional(int, dataclasses.field(default=1), fields.IntegerField, {'required': False})
        self.check_field_additional(int, dataclasses.field(default_factory=1), fields.IntegerField, {'required': False})

        # Check that Optional sets allow_null to True (#16, #38).
        self.check_field_additional(typing.Optional[int], None, fields.IntegerField, {'allow_null': True})

        # Check that Final sets exactly read_only.
        self.check_field(Final[int], fields.IntegerField, {'read_only': True})

    def test_composite(self):
        var_type = typing.TypeVar('var_type')

        self.check_field(typing.Iterable[str], custom_fields.IterableField, {'child': fields.CharField})
        self.check_field(typing.Sequence[str], custom_fields.IterableField, {'child': fields.CharField})
        self.check_field(typing.Tuple[str], custom_fields.IterableField,
                         {'child': fields.CharField, 'container': tuple})
        self.check_field(typing.Set[str], custom_fields.IterableField, {'child': fields.CharField, 'container': set})
        self.check_field(typing.List[str], custom_fields.IterableField, {'child': fields.CharField, 'container': list})
        self.check_field(typing.List[typing.Any], custom_fields.IterableField, {'container': list})
        self.check_field(typing.List[var_type], custom_fields.IterableField, {'container': list})
        self.check_field(typing.List, custom_fields.IterableField, {'container': list})
        self.check_field(list, custom_fields.IterableField, {})

        self.check_field(typing.Mapping[str, int], custom_fields.MappingField, {'child': fields.IntegerField})
        self.check_field(typing.OrderedDict[str, int], custom_fields.MappingField,
                         {'child': fields.IntegerField, 'container': collections.OrderedDict})
        self.check_field(typing.Dict[str, int], custom_fields.MappingField,
                         {'child': fields.IntegerField, 'container': dict})
        self.check_field(typing.Dict[str, typing.Any], custom_fields.MappingField, {'container': dict})
        self.check_field(typing.Dict[str, var_type], custom_fields.MappingField, {'container': dict})
        self.check_field(typing.Dict, custom_fields.MappingField, {'container': dict})
        self.check_field(dict, custom_fields.MappingField, {})

        self.check_field(typing.Optional[typing.List[str]], custom_fields.IterableField,
                         {'allow_null': True, 'child': fields.CharField, 'container': list})
        self.check_field(typing.Optional[typing.Dict[str, int]], custom_fields.MappingField,
                         {'allow_null': True, 'child': fields.IntegerField, 'container': dict})

        # check that kwargs generated for the child field are actually applied
        _, list_kwargs = self.build_typed_field(typing.List[typing.Optional[str]])
        self.assertIsInstance(list_kwargs['child'], fields.CharField)
        self.assertTrue(list_kwargs['child'].allow_null)

        _, dict_kwargs = self.build_typed_field(typing.Dict[str, Literal['a', 'b', '']])
        self.assertIsInstance(dict_kwargs['child'], fields.ChoiceField)
        self.assertDictEqual(dict_kwargs['child'].choices, {'a': 'a', 'b': 'b'})
        self.assertTrue(dict_kwargs['child'].allow_blank)

        # check that child_kwargs are applied
        _, list_kwargs = self.build_typed_field(typing.List[int],
                                                extra_kwargs={'child_kwargs': {'max_value': 5}})
        self.assertIsInstance(list_kwargs['child'], fields.IntegerField)
        self.assertEqual(list_kwargs['child'].max_value, 5)

    @unittest.skipIf(sys.version_info < (3, 9, 0), 'Python 3.9 required')
    def test_composite_pep585(self):
        self.check_field(abc.Iterable[str], custom_fields.IterableField, {'child': fields.CharField})
        self.check_field(abc.Sequence[str], custom_fields.IterableField, {'child': fields.CharField})
        self.check_field(tuple[str], custom_fields.IterableField, {'child': fields.CharField, 'container': tuple})
        self.check_field(set[str], custom_fields.IterableField, {'child': fields.CharField, 'container': set})
        self.check_field(list[str], custom_fields.IterableField, {'child': fields.CharField, 'container': list})
        self.check_field(list[typing.Any], custom_fields.IterableField, {'container': list})

        self.check_field(abc.Mapping[str, int], custom_fields.MappingField, {'child': fields.IntegerField})
        self.check_field(collections.OrderedDict[str, int], custom_fields.MappingField,
                         {'child': fields.IntegerField, 'container': collections.OrderedDict})
        self.check_field(dict[str, int], custom_fields.MappingField, {'child': fields.IntegerField, 'container': dict})
        self.check_field(dict[str, typing.Any], custom_fields.MappingField, {'container': dict})

        # check that kwargs generated for the child field are actually applied
        _, list_kwargs = self.build_typed_field(list[typing.Optional[str]])
        self.assertIsInstance(list_kwargs['child'], fields.CharField)
        self.assertTrue(list_kwargs['child'].allow_null)

        _, dict_kwargs = self.build_typed_field(dict[str, Literal['a', 'b', '']])
        self.assertIsInstance(dict_kwargs['child'], fields.ChoiceField)
        self.assertDictEqual(dict_kwargs['child'].choices, {'a': 'a', 'b': 'b'})
        self.assertTrue(dict_kwargs['child'].allow_blank)

    @unittest.skipIf(sys.version_info < (3, 10, 0), 'Python 3.10 required')
    def test_composite_pep604(self):
        self.check_field(typing.List[str] | None, custom_fields.IterableField,
                         {'child': fields.CharField, 'allow_null': True, 'container': list})
        self.check_field(typing.Dict[str, int] | None, custom_fields.MappingField,
                         {'child': fields.IntegerField, 'allow_null': True, 'container': dict})

    def test_nested(self):
        refclass = dataclasses.make_dataclass('ReferencedDataclass', [])
        self.check_field(refclass, DataclassSerializer,
                         {'dataclass': refclass, 'many': False})

        # customizing the dataclass serializer by changing the serializer_dataclass_field property
        subclassed_serializer = type('SubclassedDataclassSerializer1', (DataclassSerializer, ), {})
        setattr(DataclassSerializer, 'serializer_dataclass_field', subclassed_serializer)
        self.check_field(refclass, subclassed_serializer, {'dataclass': refclass, 'many': False})

        # customizing the dataclass serializer by putting the type into the field mapping
        subclassed_serializer = type('SubclassedDataclassSerializer2', (DataclassSerializer, ), {})
        DataclassSerializer.serializer_field_mapping[refclass] = subclassed_serializer
        self.check_field(refclass, subclassed_serializer, {'dataclass': refclass, 'many': False})

        # customizing the dataclass serializer by putting regular Field in the field mapping
        DataclassSerializer.serializer_field_mapping[refclass] = fields.CharField
        self.check_field(refclass, fields.CharField, {})

    def test_relational(self):
        django.setup()

        class Person(models.Model):
            name = models.CharField(max_length=30)

            class Meta:
                app_label = 'tests'

        self.check_field(Person, relations.PrimaryKeyRelatedField,
                         {'queryset': Person._default_manager})

        # test changing serializer_related_field, and a hyperlinked field which triggers a different code path
        DataclassSerializer.serializer_related_field = relations.HyperlinkedRelatedField
        self.check_field(Person, relations.HyperlinkedRelatedField,
                         {'queryset': Person._default_manager, 'view_name': 'person-detail'})

    def test_union(self):
        cls, kwargs = self.build_typed_field(typing.Union[str, int])
        self.assertEqual(cls, custom_fields.UnionField)
        self.assertEqual(type(kwargs['child_fields'][str]), fields.CharField)
        self.assertEqual(type(kwargs['child_fields'][int]), fields.IntegerField)

        # check that kwargs generated for the child field are applied
        _, kwargs = self.build_typed_field(typing.Union[int, Literal['a', 'b']])
        self.assertEqual(kwargs['child_fields'][Literal['a', 'b']].choices, {'a': 'a', 'b': 'b'})

        # check that child_kwargs are applied
        _, kwargs = self.build_typed_field(typing.Union[str, int], extra_kwargs={'child_kwargs': {'label': 'Test'}})
        self.assertEqual(kwargs['child_fields'][str].label, 'Test')

    def test_literal(self):
        self.check_field(Literal['a', 'b'], fields.ChoiceField,
                         {'choices': ['a', 'b'], 'allow_blank': False})
        self.check_field(Literal['a', 'b', ''], fields.ChoiceField,
                         {'choices': ['a', 'b'], 'allow_blank': True})
        self.check_field(Literal['a', 'b', None], fields.ChoiceField,
                         {'choices': ['a', 'b'], 'allow_blank': False, 'allow_null': True})
        self.check_field(Literal['a', 'b', '', None], fields.ChoiceField,
                         {'choices': ['a', 'b'], 'allow_blank': True, 'allow_null': True})

    def test_enum(self):
        class Color(enum.Enum):
            RED = enum.auto()
            GREEN = enum.auto()
            BLUE = enum.auto()

        self.check_field(Color, custom_fields.EnumField, {'enum_class': Color})

    def test_standard_primitives(self):
        self.check_field(int, fields.IntegerField)
        self.check_field(float, fields.FloatField)
        self.check_field(bool, fields.BooleanField)
        self.check_field(str, fields.CharField)

    def test_standard_variable(self):
        var_str = typing.TypeVar('var_str', bound=str)
        var_any = typing.TypeVar('var_any')

        self.check_field(var_str, fields.CharField)

        with self.assertRaises(NotImplementedError):
            self.build_typed_field(var_any)

    def test_standard_decimal(self):
        self.check_field(decimal.Decimal, custom_fields.DefaultDecimalField)

    def test_standard_dates(self):
        self.check_field(datetime.date, fields.DateField)
        self.check_field(datetime.datetime, fields.DateTimeField)
        self.check_field(datetime.time, fields.TimeField)
        self.check_field(datetime.timedelta, fields.DurationField)

    def test_standard_uuid(self):
        self.check_field(uuid.UUID, fields.UUIDField)

    def test_standard_subclass(self):
        custom_str = type('custom_str', (str, ), {})
        custom_date = type('custom_date', (datetime.date, ), {})
        self.check_field(custom_str, fields.CharField)
        self.check_field(custom_date, fields.DateField)

    def test_standard_error(self):
        msg = ("Automatic serializer field deduction not supported for field 'test_field' on 'TestDataclass' "
               "of type '<class 'complex'>' (during search for field of type '<class 'complex'>').")
        with self.assertRaisesRegex(NotImplementedError, re.escape(msg)):
            self.build_typed_field(complex)

        # Check _SpecialForm types that don't have an __mro__attribute (#6)
        msg = ("Automatic serializer field deduction not supported for field 'test_field' on 'TestDataclass' "
               "of type 'typing.Any' (during search for field of type 'typing.Any').")
        with self.assertRaisesRegex(NotImplementedError, re.escape(msg)):
            self.build_typed_field(typing.Any)



from __future__ import annotations

import sys
from dataclasses import Field
from typing import Any, ClassVar, Dict

# Alias these types into this module for ease of use elsewhere.
if sys.version_info >= (3, 8):
    from typing import Final, Literal, Protocol
else:
    from typing_extensions import Final, Literal, Protocol


# As dataclasses don't have a baseclass, we typehint them using a protocol matching the injected `__dataclass_fields__`
# attribute. This is also what the Python stdlib typeshed does:
# https://github.com/python/typeshed/blob/ced150a7e8edafe98ffceb3365fe66f800bc6780/stdlib/_typeshed/__init__.pyi#L309-L314
# Requires MyPy 0.990+
class Dataclass(Protocol):
    __dataclass_fields__: ClassVar[Dict[str, Field[Any]]]
